use crate::bytes::Bytes;

pub use openssl::symm::{Cipher, Crypter, Mode};

pub struct Aes {
    key: Vec<u8>,
    pad: bool,
}

impl Aes {
    const BLOCK_SIZE: usize = 16;

    pub fn new(key: &[u8], pad: bool) -> Option<Aes> {
        let len = key.len();
        if len == 16 || len == 24 || len == 32 {
            return Some(Aes {
                key: key.to_vec(),
                pad,
            });
        };
        None
    }

    pub fn cbc(&self, data: Vec<u8>, iv: &[u8], mode: Mode) -> Option<Vec<u8>> {
        if (!self.pad) && (data.len() % Self::BLOCK_SIZE != 0) {
            return None;
        }
        let mut data = self.maybe_pad(data, mode);
        let mut iv_ = [0u8; Self::BLOCK_SIZE];
        let mut output_buffer = [0u8; 2 * Self::BLOCK_SIZE];

        iv_.clone_from_slice(iv);
        let mut iv = iv_;

        let mut crypter = self.crypter(mode);
        for chunk in data.chunks_exact_mut(Self::BLOCK_SIZE) {
            match mode {
                Mode::Decrypt => {
                    let mut old_chunk = [0u8; Self::BLOCK_SIZE];
                    old_chunk.copy_from_slice(chunk);

                    Self::update(&mut crypter, chunk, &mut output_buffer);
                    Self::xor_into(&output_buffer, &iv, chunk);

                    iv.clone_from_slice(&old_chunk);
                }
                Mode::Encrypt => {
                    Self::xor(chunk, &iv);

                    Self::update(&mut crypter, chunk, &mut output_buffer);
                    chunk.copy_from_slice(&output_buffer[0..Self::BLOCK_SIZE]);

                    iv.copy_from_slice(chunk);
                }
            }
        }
        let data = self.maybe_unpad(data, mode);
        Some(data.clone())
    }

    pub fn ecb(&self, mut data: Vec<u8>, mode: Mode) -> Option<Vec<u8>> {
        if (!self.pad) && (data.len() % Self::BLOCK_SIZE != 0) {
            return None;
        }

        let mut crypter = self.crypter(mode);
        let mut output_buffer = [0u8; 2 * Self::BLOCK_SIZE];
        let mut data = self.maybe_pad(data, mode);
        for chunk in data.chunks_exact_mut(Self::BLOCK_SIZE) {
            Self::update(&mut crypter, chunk, &mut output_buffer);
            chunk.copy_from_slice(&output_buffer[0..Self::BLOCK_SIZE]);
        }

        let data = self.maybe_unpad(data, mode);
        Some(data)
    }

    fn cipher(&self) -> Cipher {
        match self.key.len() {
            16 => Some(Cipher::aes_128_ecb()),
            24 => Some(Cipher::aes_192_ecb()),
            32 => Some(Cipher::aes_256_ecb()),
            _ => None,
        }
        .unwrap()
    }

    fn crypter(&self, mode: Mode) -> Crypter {
        let cipher = self.cipher();
        let mut crypter = Crypter::new(cipher, mode, &self.key, None).unwrap();
        crypter.pad(false);
        crypter
    }

    fn maybe_pad(&self, data: Vec<u8>, mode: Mode) -> Vec<u8> {
        match (self.pad, mode) {
            (true, Mode::Encrypt) => Bytes { data }.pad_to_multiple(Self::BLOCK_SIZE).into(),
            _ => data,
        }
    }

    fn maybe_unpad(&self, data: Vec<u8>, mode: Mode) -> Vec<u8> {
        match (self.pad, mode) {
            (true, Mode::Decrypt) => Bytes { data }.unpad().unwrap().into(),
            _ => data,
        }
    }


    fn update(crypter: &mut Crypter, chunk: &[u8], output: &mut [u8]) {
        assert!(
            Self::BLOCK_SIZE
                == crypter
                    .update(chunk, output)
                    .expect("Error while updating crypter")
        );
    }

    fn xor(a: &mut [u8], b: &[u8]) {
        for (l, r) in a.iter_mut().zip(b.iter()) {
            *l = *l ^ r;
        }
    }

    fn xor_into(a: &[u8], b: &[u8], c: &mut [u8]) {
        for ((t, l), r) in c.iter_mut().zip(a.iter()).zip(b.iter()) {
            *t = l ^ r;
        }
    }
}
